package mmc;

/**
 * medium definition
 */

public class Medium extends PhysicsModel{

    public int num;                // number of components
    public double[] Z;             // nuclius charge
    public double[] A;             // atomic number
    public double[] n;             // number of atoms in a molecule
    public double totZ;            // sum of charges of all nuclei

    public double ZA;              // <Z/A>
    public double I;               // ionization potential [eV]
    public double C1, C, a;        // ionization formula constants
    public double m, X0, X1;       // ionization formula constants (continued)
    public double r;               // refraction index

    public double B[];             // radiation logarithm constant B
    public double P[];             // radiation logarithm constant bPrime

    public double Ro;              // mass density [g/cm3]
    public double No;              // molecule density [number/cm3]
    public double M[];             // average nucleon weight in a nucleus [MeV]
    public String E[];             // element name
    public String name;            // medium name

    protected double ecut;         // cutoff energy [MeV]
    protected double vcut;         // relative cutoff energy
    private double vCut;           // relative cutoff energy - call setCut(E) to set this


    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize medium by its name and proposed values of energy cut (ecut) and fractional energy cut (vcut).
     * The cuts ecut and vcut to be used must satisfy ecut&gt;0 and 0&lt;=vcut&lt;=1; If both values satisfy these
     * inequalities, the lower from E*vcut and ecut will be used. If only one value satisfies the inequalities,
     * only that one value is used. If both values are outside these intervals, vcut=1 is assumed.
     */

    public Medium(String w, double ecut, double vcut){
	name=w;
	if(w.equalsIgnoreCase("water")) initWater();
	else if(w.equalsIgnoreCase("ice")) initIce();
	else if(w.equalsIgnoreCase("standard rock")) initStandardrock();
	else if(w.equalsIgnoreCase("frejus rock")) initFrejusrock();
	else if(w.equalsIgnoreCase("iron")) initIron();
	else if(w.equalsIgnoreCase("hydrogen")) initHydrogen();
	else if(w.equalsIgnoreCase("uranium")) initUranium();
	else {
	    System.err.println("Warning (in Medium/Medium): unknown medium: \""+w+"\", defaulting to water");
	    name="water";
	    initWater();
	}
	this.ecut=ecut;
	this.vcut=vcut;
	vCut=1.;
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialization of arrays
     */

    private void inita(int i){
	num=i;
	Z = new double[num];
	A = new double[num];
	n = new double[num];
	B = new double[num];
	P = new double[num];
	M = new double[num];
	E = new String[num];
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialization code, common to all media
     */

    private void initr(){
	int i;
	double aux1=0, aux2=0;
	for(i=0;i<num;i++){
	    aux1+=n[i]*Z[i];
	    aux2+=n[i]*A[i];
	    B[i]=elB(i);
	    P[i]=elP(i);
	    M[i]=(Z[i]*Mp+(A[i]-Z[i])*Mn)/A[i];
	}
	totZ=aux1;
	ZA=aux1/aux2;
	No=Ro*Na/aux2;
	C1=4.606;
	r=1.31;                         // only for ice - change if needed
	ecut=Me/Math.sqrt(1-1/(r*r));   // in order to emit cherenkov photons
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * set the value of radiation logarithm constant B
     */

    private double elB(int i){
	int z=(int)Math.round(Z[i]);
	switch(z){
	case 1: return 202.4;
	case 2: return 151.9;
	case 3: return 159.8;
	case 4: return 172.3;
	case 5: return 177.8;
	case 6: return 178.3;
	case 7: return 176.5;
	case 8: return 173.3;
	case 9: return 169.9;
	case 10: return 165.7;
	case 11: return 165.7;
	case 12: return 167.0;
	case 13: return 169.0;
	case 14: return 170.7;
	case 15: return 172.1;
	case 16: return 173.3;
	case 17: return 174.2;
	case 18: return 174.7;
	case 19: return 176.6;
	case 20: return 178.5;
	case 21: return 179.2;
	case 22: return 179.8;
	case 23: return 179.9;
	case 26: return 179.0;
	case 29: return 176.5;
	case 32: return 176.8;
	case 35: return 177.6;
	case 42: return 180.7;
	case 50: return 182.7;
	case 53: return 183.9;
	case 74: return 183.4;
	case 82: return 184.1;
	case 92: return 186.6;
	default: return 182.7;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * set the value of radiation logarithm constant bPrime
     */

    private double elP(int i){
	int z=(int)Math.round(Z[i]);
	switch(z){
	case 1: return 446;
	default: return 1429;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize water
     */

    private void initWater(){
	inita(2);
	E[0]="H";
	E[1]="O";
	Z[0]=1; // H
	Z[1]=8; // O
	A[0]=1.00794;
	A[1]=15.9994;
	n[0]=2;
	n[1]=1;
	I=75.0;
	C=-3.5017;
	a=0.09116;
	m=3.477;
	X0=0.240;
	X1=2.8004;
	Ro=1.000;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize ice
     */

    private void initIce(){
	inita(2);
	E[0]="H";
	E[1]="O";
	Z[0]=1; // H
	Z[1]=8; // O
	A[0]=1.00794;
	A[1]=15.9994;
	n[0]=2;
	n[1]=1;
	I=75.0;
	C=-3.5017;
	a=0.09116;
	m=3.477;
	X0=0.240;
	X1=2.8004;
	Ro=0.917;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize standard rock
     */

    private void initStandardrock(){
	inita(1);
	E[0]="Standard Rock";
	Z[0]=11; // Ionization potential and density corrections
	A[0]=22; // are close to those of calcium carbonate
	n[0]=1;
	I=136.4;
	C=-3.774;
	a=0.083;
	m=3.412;
	X0=0.049;
	X1=3.055;
	Ro=2.650;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize standard rock
     */

    private void initFrejusrock(){
	inita(1);
	E[0]="Frejus Rock";
	Z[0]=10.12;
	A[0]=20.34;
	n[0]=1;
	I=149.0;
	C=-5.053;
	a=0.078;
	m=3.645;
	X0=0.288;
	X1=3.196;
	Ro=2.740;
	initr();
    }


    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize iron
     */

    private void initIron(){
	inita(1);
	E[0]="Fe";
	Z[0]=26;
	A[0]=55.845;
	n[0]=1;
	I=286.0;
	C=-4.291;
	a=0.147;
	m=2.963;
	X0=-0.001;
	X1=3.153;
	Ro=7.874;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize hydrogen
     */

    private void initHydrogen(){
	inita(1);
	E[0]="H";
	Z[0]=1;
	A[0]=1.00794;
	n[0]=2;
	I=21.8;
	C=-3.263;
	a=0.135;
	m=5.625;
	X0=0.476;
	X1=1.922;
	Ro=0.063;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize hydrogen
     */

    private void initUranium(){
	inita(1);
	E[0]="U";
	Z[0]=92;
	A[0]=238.0289;
	n[0]=1;
	I=890.0;
	C=-5.869;
	a=0.197;
	m=2.817;
	X0=0.226;
	X1=3.372;
	Ro=18.950;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * return the value of the fractional energy cut, as described in the help for the constructor
     */

    public double vCut(double E){
	double aux;
	aux=ecut/E;
	if(ecut>0){
	    if(vcut>0 && vcut<=1){
		if(aux<vcut) vCut=aux;
		else vCut=vcut;
	    }
	    else vCut=aux;
	}
	else{
	    if(vcut>0 && vcut<=1) vCut=vcut;
	    else vCut=1.;
	}
	return vCut;
    }

}
