package mmc;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.text.DecimalFormat;

public class Output{

    final public static String version="Muon Propagation Monte Carlo Written in Java, v 0.11";

    public static boolean DEBUG=false;
    public String history;
    public int gen, gens;

    protected int HIST=-1;
    protected Particle p;

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize all classes necessary for propagation
     */

    public Output(Particle p){
	this.p = p;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize for no history output - no need to call this function
     */

    public void init(String type){
	HIST=0;
	p.location(type, 0, 0, 0, 0, 0, 0);
	history="";
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize for regular history output
     */

    public void initDefault(String type){
	HIST=1;
	p.location(type, 0, 0, 0, 0, 0, 0);
	history="";
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize for regular history output
     */

    public void initDefault(String type, double time, double x, double y, double z, double theta, double phi){
	HIST=3;
	p.location(type, time, x, y, z, theta, phi);
	history="";
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize for F2000 history output
     */

    public void initF2000(int igen, String type, double time, double x, double y, double z, double theta, double phi){
	HIST=2;
	gen=igen;
	gens=gen;
	p.location(type, time, x, y, z, theta, phi);
	history="";
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * print output information in the format specified by HIST
     */

    protected void output(int wint, String comp, double de, double ef){
	String str;
	if(DEBUG || (HIST==1)){
	    str=" * ";
	    if(wint==1) str+="decayed";
	    else if(wint==2) str+="ionized by "+comp;
	    else if(wint==3) str+="bremsed by "+comp;
	    else if(wint==4) str+="photoed by "+comp;
	    else if(wint==5) str+="epaired by "+comp;
	    str+=" ... lost "+f(de)+" MeV, ef = "+f(ef)+" MeV, xf = "+f(p.r)+" cm\n";
	    if(HIST==1) history+=str;
	    if(DEBUG) System.err.print(str);
	}
	if(HIST==2){
	    if(wint==1){
		if("mu+".equals(p.name)) str="e+";
		else if("mu-".equals(p.name)) str="e-";
		else str="e";
	    }
	    else if(wint==2) str="delta";
	    else if(wint==3) str="brems";
	    else if(wint==4) str="munu";
	    else if(wint==5) str="epair";
	    else str="";
	    if(str!=""){
		gens++;
		str="TR "+gens+" "+gen+" "+str+" "+f(p.x*1.e-2)+" "+f(p.y*1.e-2)+" "+f(p.z*1.e-2);
		str+=" "+f(180-p.theta)+" "+f(p.phi<180?p.phi+180:p.phi-180)+" 0 "+f(de*1.e-3)+" "+f(p.t*1.e9)+"\n";
		history+=str;
	    }
	}
    }

    //----------------------------------------------------------------------------------------------------//

    protected static LineNumberReader in=null;
    protected static PrintWriter out=null;

    //----------------------------------------------------------------------------------------------------//

    /**
     * read one number from the table
     */

    protected static double read(){
	try{
	    String str=in.readLine();
	    if(str==null) throw new mmcException("parametrization table premature eof");
	    return Double.parseDouble(str);
	}catch(Exception error){
	    System.err.println("***WARNING***: function failed: "+error.toString());
	    close();
	    throw new mmcException("parametrization table read error");
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * write one number to the table
     */

    protected static void write(double g){
	try{
	    out.println(g);
	}catch(Exception error){
	    System.err.println("***WARNING***: function failed: "+error.toString());
	    return;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * open the table
     */

    protected static void open(String name){
	try{
	    if((new File(name)).exists()){
		in = new LineNumberReader(new FileReader(name));
		out = null;
		System.err.println("Parametrization tables will be read in from the file "+name);
	    }
	    else{
		in = null;
		out = new PrintWriter(new FileWriter(name));
		System.err.println("Parametrization tables will be saved to the file "+name);
	    }
	}catch(Exception error){
	    in = null;
	    out = null;
	    System.err.println("Files cannot be used: "+error.toString());
	    return;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * close the table
     */

    protected static void close(){
	if(out!=null) out.close();
	in = null;
	out = null;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * delete the file
     */

    protected static void delete(String name){
	try{
	    System.err.println("... Deleting corrupt file "+name);
	    (new File(name)).delete();
	}catch(Exception error){
	    System.err.println("Cannot delete file: "+error.toString());
	    return;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    private static DecimalFormat df1 = new DecimalFormat(".#######");
    private static DecimalFormat df2 = new DecimalFormat("0.######");
    private static DecimalFormat df3 = new DecimalFormat("#0.#####");
    private static DecimalFormat df4 = new DecimalFormat("##0.####");
    private static DecimalFormat df5 = new DecimalFormat("###0.###");
    private static DecimalFormat df6 = new DecimalFormat("####0.##");
    private static DecimalFormat df7 = new DecimalFormat("#####0.#");
    private static DecimalFormat df8 = new DecimalFormat("######0");
    private static DecimalFormat df9 = new DecimalFormat("0.######E0");

    /**
     * Format the double
     */

    public static String f(double d){
	double aux;
	if(d==0) return "0";
	aux=Math.abs(d);
	if(aux<0.1) return df9.format(d);
	else if(aux<1) return df1.format(d);
	else if(aux<10) return df2.format(d);
	else if(aux<100) return df3.format(d);
	else if(aux<1e3) return df4.format(d);
	else if(aux<1e4) return df5.format(d);
	else if(aux<1e5) return df6.format(d);
	else if(aux<1e6) return df7.format(d);
	else if(aux<1e7) return df8.format(d);
	else return df9.format(d);
    }

}
