package mmc;

/**
 * particle definition (muon)
 */

public class Particle extends PhysicsModel{

    public double r=0, x=0, y=0, z=0, t=0;             // coordinates [cm] and age [sec]
    public double theta=0, phi=0;                      // zenith and azimuth of the momentum in [deg]
    public double costh=1, sinth=0, cosph=1, sinph=0;  // cos and sin of theta and phi
    public double p=0, p2=0, e=0;                      // momentum, momentum square and energy in [MeV]
    public double m=105.658389, l=2.19703e-6, c=1;     // mass [MeV], lifetime [sec] and charge
    public String name="mu";                           // name of the particle
    public double low=m;                               // energy below which the particle is lost [MeV]

    private Propagate pr;
    private Integral I;

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize particle
     */

    public Particle(Propagate pr){
	e=m;
	this.pr=pr;
 	I = new Integral(5, 40, 1.e-5);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize the location and direction of the particle, time in sec, x, y, z in cm, theta and phi in deg
     */

    public void location(String type, double time, double x, double y, double z, double theta, double phi){
	name=type;
	r=0;
	t=time;
	this.x=x;
	this.y=y;
	this.z=z;
	this.theta=theta;
	this.phi=phi;
	theta*=(Pi/180);
	phi*=(Pi/180);
	costh=Math.cos(theta);
	sinth=Math.sin(theta);
	cosph=Math.cos(phi);
	sinph=Math.sin(phi);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * advances the particle by the given distance
     */

    public void advance(double dr, double ei, double ef){
	r+=dr;
	if(pr.exactTime) t+=getdt(ei, ef);
	else t+=dr/C;
	x+=sinth*cosph*dr;
	y+=sinth*sinph*dr;
	z+=costh*dr;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * sets the energy of the particle
     */

    public void setEnergy(double e){
	this.e=e;
	p2=e*e-m*m;
	p=Math.sqrt(Math.max(p2, 0));
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * function for time delta calculation - interface to Integral
     */

    public double function(double E){
	final boolean DEBUG=false;
	double aux;
	aux=pr.s.function(E);
	aux*=e/(p*C);
	if(DEBUG) System.err.println(" # "+Output.f(e)+" \t "+Output.f(aux));
	return aux;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * time delta, corresponding to the given propagation distance
     */

    public double getdt(double ei, double ef){
	if(jt) return J.interpolate(ei)-J.interpolate(ef);
	else return I.integrateWithLog(ei, ef, this);
    }

    //----------------------------------------------------------------------------------------------------//

    public Interpolate J;
    public boolean jt=false;

    /**
     * 2d parametrization - interface to Interpolate
     */

    public double functionInt(double e){
	return I.integrateWithLog(e, low, this);
    }

}
