package mmc;

/**
 * class contains functions necessary for calculation of photonuclear losses
 */

public class Photonuclear extends CrossSections{

    public PhotoContinuous c;
    public PhotoStochastic s;

    protected double vMax, vUp, vMin;
    protected Photonuclear photo;

    //----------------------------------------------------------------------------------------------------//

    /**
     * creates internal references to p and m, to be called from subclasses
     */

    public Photonuclear(Photonuclear cros){
	p=cros.p;
	m=cros.m;
	this.cros=cros.cros;
	photo=cros;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initializes subclasses and creates internal references to p and m
     */

    public Photonuclear(CrossSections cros){
	super(cros);
	photo=this;
	c = new PhotoContinuous(this);
	s = new PhotoStochastic(this);
	setMeasured();
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * call before using the photonuclear functions to set the component of the primary
     */

    public void setEnergy(int i){
	double aux;
	cros.component=i;
	vMin=(Mpi+(Mpi*Mpi)/(2*m.M[i]))/p.e;
	aux=p.m/m.M[i];
	vMax=1-0.5*m.M[i]*(1+aux*aux)/p.e;
	vMax=Math.min(vMax, 1-p.m/p.e);
	if(vMax<vMin) vMax=vMin;
	vUp=Math.min(vMax, m.vCut(p.e));
	if(vUp<vMin) vUp=vMin;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * this is what the photonuclear interaction cross section is equal to
     */

    public double photoN(double v, int i){
	final double m1=0.54, m2=1.80;
	double nu, sgn, aux, k, G, t;

	nu=v*p.e;
	nu*=1.e-3;

	if(nu<=200.){
	    if(measured) sgn=measuredSgN(nu);
	    else{
		if(nu<=17.) sgn=96.1+82./Math.sqrt(nu);
		else{ aux=Math.log(0.0213*nu); sgn=114.3+1.647*aux*aux; }
	    }
	}
	else sgn=49.2+11.1*Math.log(nu)+151.8/Math.sqrt(nu);

	k=1-2/v+2/(v*v);

	if(m.Z[i]==1) G=1;
	else {
	    aux=0.00282*Math.pow(m.A[i], 1./3)*sgn;
	    G=(3/aux)*(0.5+((1+aux)*Math.exp(-aux)-1)/(aux*aux));
	}
	G*=3;

	aux=v*p.m*1.e-3;
	t=aux*aux/(1-v);
	sgn*=1.e-30;

	aux=p.m*1.e-3;
	aux=2*aux*aux/t;
	aux=G*(k*Math.log(1+m1/t)-(k*m1)/(m1+t)-aux)+(k*Math.log(1+m2/t)-aux)+0.25*aux*((G*m1)/(m1+t)+(m2/t)*Math.log(1+t/m2));
	aux*=Alpha/(8*Pi)*m.A[i]*sgn*v;

	return m.No*m.n[i]*aux;
    }

    //----------------------------------------------------------------------------------------------------//

    private Interpolate J;
    public boolean measured=false;

    //----------------------------------------------------------------------------------------------------//

    /*
     * call this to replace photon-nucleon cross section approximation formula with
     * a more exact at low energies parametrization of the experimental data
     */

    private void setMeasured(){
	final double x[]={0, 0.1, 0.144544, 0.20893, 0.301995, 0.436516, 0.630957, 0.912011, 1.31826, 1.90546, 2.75423, 3.98107, 5.7544, 8.31764, 12.0226, 17.378, 25.1189, 36.3078, 52.4807, 75.8577, 109.648, 158.489, 229.087, 331.131, 478.63, 691.831, 1000, 1445.44, 2089.3, 3019.95, 4365.16, 6309.58, 9120.12, 13182.6, 19054.6, 27542.3, 39810.8, 57544, 83176.4, 120226, 173780, 251188, 363078, 524807, 758576, 1.09648e+06, 1.58489e+06, 2.29086e+06, 3.3113e+06, 4.78628e+06, 6.91828e+06, 9.99996e+06};
	final double y[]={0, 0.0666667, 0.0963626, 159.74, 508.103, 215.77, 236.403, 201.919, 151.381, 145.407, 132.096, 128.546, 125.046, 121.863, 119.16, 117.022, 115.496, 114.607, 114.368, 114.786, 115.864, 117.606, 120.011, 123.08, 126.815, 131.214, 136.278, 142.007, 148.401, 155.46, 163.185, 171.574, 180.628, 190.348, 200.732, 211.782, 223.497, 235.876, 248.921, 262.631, 277.006, 292.046, 307.751, 324.121, 341.157, 358.857, 377.222, 396.253, 415.948, 436.309, 457.334, 479.025};
	J = new Interpolate(x, y, 4, false, false);
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * returns measured value of the photon nucleon cross section
     */

    private double measuredSgN(double e){
	return J.interpolateArray(e);
    }

}
