package mmc;

public class Propagate extends PhysicsModel{

    public Particle p;
    public Medium m;
    public CrossSections s;
    public Output o;

    public boolean sdec=false;
    public boolean exactTime=false;
    public boolean contiCorr=false;

    private Integral I;
    private StandardNormal S;
    private double decayS, ionizS, bremsS, epairS, photoS, totalS;

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize all classes necessary for propagation. String w containes the description of the medium.
     * For the definition of ecut and vcut see help for the class Medium constructor. To enable
     * parametrization routines, call this.interpolate("all") after this class is created.
     */

    public Propagate(String w, double ecut, double vcut){
	p = new Particle(this);
	m = new Medium(w, ecut, vcut);
	s = new CrossSections(p, m);
 	I = new Integral(5, 40, 1.e-5);
	S = new StandardNormal();
	o = new Output(p);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * Propagates the particle of initial energy e to the z-component of the distance z. Returns the final energy
     * if the particle has survived or the track length to the point of disappearance with a minus sign otherwise.
     */

    public double propagateTo(double z, double e){
	int wint;
	boolean DEBUG, flag;
	double ei, ef=0, aux=0, ini, dr;
	double rnd, rnd1, rnd2, rnd3, rndMin, rndTot;

	DEBUG=Output.DEBUG;
	if(o.HIST==-1) o.init("mu");

	ei=e;
	if(e<=p.low) flag=false;
	else flag=true;
	if(DEBUG) System.err.println("\nPropagating "+p.name+" of energy "+ei);

	while(flag){
	    rnd=-Math.log(Math.random());
	    if(DEBUG) System.err.println("1. solving the tracking integral ...");
	    rndMin=getpr(ei, rnd);
	    if(DEBUG) System.err.println(" * rnd = "+Output.f(rnd)+" rndMin = "+Output.f(rndMin));

	    if(DEBUG) System.err.println("2. evaluating the energy loss ...");
	    if(rnd>=rndMin || rndMin<=0) ef=p.low;
	    else ef=getef(ei, rndMin, rnd);

	    if(DEBUG) System.err.println("3. calculating the displacement ...");
	    ini=s.getini(ei);
	    dr=s.getdx(ei, ef, ini, (z-p.z)/p.costh);
	    if(dr<(z-p.z)/p.costh){
		if(DEBUG) System.err.println("4. calculating the local time ...");
		p.advance(dr, ei, ef);
	    }
	    else{
		if(DEBUG) System.err.println("4. getting the final energy ...");
		ef=s.getef(ei, ini, (z-p.z)/p.costh);
		if(DEBUG) System.err.println("5. calculating the local time ...");
		p.advance((z-p.z)/p.costh, ei, ef);
	    }
	    if(Math.abs(p.z-z)<Math.abs(z)*computerPrecision) p.z=z;  // computer precision control
	    if(contiCorr) if(ef!=p.low) ef=S.sndrn(Math.random(), ef, Math.sqrt(ei*ei-ef*ef)*m.vCut((ef+ei)/2), p.low, ei, false);
	    if(ef==p.low || p.z==z) break;

	    if(DEBUG) System.err.println("5. choosing the cross section ...");
	    rnd1=Math.random();
	    rnd2=Math.random();
	    rnd3=Math.random();
	    p.setEnergy(ef);
	    decayS=s.d.decay();
	    ionizS=s.i.s.dNdx(rnd2);
	    bremsS=s.b.s.dNdx(rnd2);
	    photoS=s.n.s.dNdx(rnd2);
	    epairS=s.e.s.dNdx(rnd2);
	    totalS=decayS+ionizS+bremsS+photoS+epairS;
	    rndTot=rnd1*totalS;
	    if(DEBUG) System.err.println(" * rnd = "+Output.f(Math.exp(-rnd))+" rnd1 = "+Output.f(rnd1)+
					 " rnd2 = "+Output.f(rnd2)+" rnd3 = "+Output.f(rnd3));
	    if(DEBUG) System.err.println(" * ioniz = "+Output.f(ionizS)+" brems = "+Output.f(bremsS)+
					 " photo = "+Output.f(photoS)+" epair = "+Output.f(epairS));

	    if(decayS>rndTot){ aux=s.d.e(rnd2, rnd3); ef=0; wint=1; }
	    else if(decayS+ionizS>rndTot){ aux=s.i.s.e(rnd3); ef-=aux; wint=2; }
	    else if(decayS+ionizS+bremsS>rndTot){ aux=s.b.s.e(rnd3); ef-=aux; wint=3; }
	    else if(decayS+ionizS+bremsS+photoS>rndTot){ aux=s.n.s.e(rnd3); ef-=aux; wint=4; }
	    else{ aux=s.e.s.e(rnd3); wint=5; }

	    o.output(wint, m.E[s.component], aux, ef);

	    if(ef<=p.low) break;
	    ei=ef;
	}

	if(sdec) if(p.z!=z && ef!=0){
	    p.setEnergy(p.m);
	    p.t+=-p.l*Math.log(Math.random());
	    aux=s.d.e(Math.random(), 0.5);
	    ef=0;
	    o.output(1, "decayed", aux, ef);
	}
	p.setEnergy(ef);  // to remember final state of the particle
	o.HIST=-1;  // to make sure user resets particle properties

	if(p.z==z){
	    if(DEBUG) System.err.println("Particle reached the border with energy ef = "+Output.f(ef)+" MeV");
	    return ef;
	}
	else{
	    if(DEBUG) System.err.println("Particle disappeared at zf = "+Output.f(p.z)+" cm");
	    return -p.r;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    private double ec;

    /**
     * Propagates the particle of initial energy e to the z-component of the distance z. Returns the final energy
     * if the particle has survived or the track length to the point of disappearance with a minus sign otherwise.
     * Also calculates particle energy at point zc. Call getPropEc() to get this energy.
     */

    public double propagateTo(double z, double e, double zc){
	int HIST;
	double result;
	HIST=o.HIST;
	ec=propagateTo(zc, e);
	if(ec>0){
	    o.HIST=HIST;
	    result=propagateTo(z, ec);
	}
	else result=ec;
	return result;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * returns the particle energy at zc if the particle has survived or the
     * distance from the point of decay to zc with a minus sign otherwise.
     */

    public double getPropEc(){
	return ec;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * function for energy range calculation - interface to Integral
     */

    public double function(double E){
	final boolean DEBUG=false;
	double aux;
	aux=s.function(E);
	decayS=s.d.decay();
	if(DEBUG) System.err.print(" + "+Output.f(p.e));
	ionizS=s.i.s.dNdx();
	if(DEBUG) System.err.print(" \t "+Output.f(ionizS));
	bremsS=s.b.s.dNdx();
	if(DEBUG) System.err.print(" \t "+Output.f(bremsS));
	photoS=s.n.s.dNdx();
	if(DEBUG) System.err.print(" \t "+Output.f(photoS));
	epairS=s.e.s.dNdx();
	if(DEBUG) System.err.println(" \t "+Output.f(epairS));
	totalS=decayS+ionizS+bremsS+photoS+epairS;
	return aux*totalS;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * call this routine to enable interpolations. To enable everything, set w="all"
     */

    public void interpolate(String w){
	int i;
	double e_hi=1.e14;
	double e_low=p.low;

	System.err.println("Parametrizations apply in the energy range from "+Output.f(e_low)+" MeV to "+Output.f(e_hi)+" MeV");
	if(w.indexOf("all")!=-1){
	    w+=" crs trackE trackX";
	    if(exactTime) w+=" trackT";
	    if(contiCorr) w+=" contiC";
	}
	if(w.indexOf("crs")!=-1) w+=" ionizE ionizS bremsE bremsS photoE photoS epairE epairP epairS";

	s.i.c.jt=false;
	if(w.indexOf("ionizE")!=-1){
	    System.err.print("Parametrizing ionizE ... ");
	    s.i.c.J = new Interpolate(100, e_low, e_hi, s.i.c, 5, true, false, true, 5, false, false, true);
	    s.i.c.jt=true;
	    System.err.println("done");
	}

	s.i.s.jt=false;
	if(w.indexOf("ionizS")!=-1){
	    System.err.print("Parametrizing ionizS ... ");
	    s.i.s.J = new Interpolate(100, e_low, e_hi, 100, 0, 1, s.i.s, 5, false, false, true, 5, false, false, false, 5, true, false, false);
	    s.i.s.jt=true;
	    System.err.println("done");
	}

	s.b.c.jt=false;
	if(w.indexOf("bremsE")!=-1){
	    System.err.print("Parametrizing bremsE ... ");
	    s.b.c.J = new Interpolate(100, e_low, e_hi, s.b.c, 5, true, false, true, 5, false, false, false);
	    s.b.c.jt=true;
	    System.err.println("done");
	}

	s.b.s.jt=false;
	if(w.indexOf("bremsS")!=-1){
	    System.err.print("Parametrizing bremsS ... ");
	    s.b.s.J = new Interpolate[m.num];
	    for(i=0; i<m.num; i++){
		s.component=i;
		s.b.s.J[i] = new Interpolate(100, e_low, e_hi, 100, 0, 1, s.b.s, 5, false, false, true, 5, false, false, false, 5, true, false, false);
	    }
	    s.b.s.jt=true;
	    System.err.println("done");
	}

	s.n.c.jt=false;
	if(w.indexOf("photoE")!=-1){
	    System.err.print("Parametrizing photoE ... ");
	    s.n.c.J = new Interpolate(100, e_low, e_hi, s.n.c, 5, true, false, true, 5, false, false, false);
	    s.n.c.jt=true;
	    System.err.println("done");
	}

	s.n.s.jt=false;
	if(w.indexOf("photoS")!=-1){
	    System.err.print("Parametrizing photoS ... ");
	    s.n.s.J = new Interpolate[m.num];
	    for(i=0; i<m.num; i++){
		s.component=i;
		s.n.s.J[i] = new Interpolate(100, e_low, e_hi, 100, 0, 1, s.n.s, 5, false, false, true, 5, false, false, false, 5, true, false, false);
	    }
	    s.n.s.jt=true;
	    System.err.println("done");
	}

	s.e.jt=false;
	if(w.indexOf("epairP")!=-1){
	    System.err.print("Parametrizing epairP ... ");
	    s.e.J = new Interpolate[m.num];
	    for(i=0; i<m.num; i++){
		s.component=i;
		s.e.J[i] = new Interpolate(100, e_low, e_hi, 100, 0., 1., s.e, 5, false, false, true, 5, false, false, false, 5, false, false, false);
	    }
	    s.e.jt=true;
	    System.err.println("done");
	}

	s.e.c.jt=false;
	if(w.indexOf("epairE")!=-1){
	    System.err.print("Parametrizing epairE ... ");
	    s.e.c.J = new Interpolate(100, e_low, e_hi, s.e.c, 5, true, false, true, 5, false, false, false);
	    s.e.c.jt=true;
	    System.err.println("done");
	}

	s.e.s.jt=false;
	if(w.indexOf("epairS")!=-1){
	    System.err.print("Parametrizing epairS ... ");
	    s.e.s.J = new Interpolate[m.num];
	    for(i=0; i<m.num; i++){
		s.component=i;
		s.e.s.J[i] = new Interpolate(100, e_low, e_hi, 100, 0, 1, s.e.s, 5, false, false, true, 5, false, false, false, 5, true, false, false);
	    }
	    s.e.s.jt=true;
	    System.err.println("done");
	}

	this.jt=false;
	if(w.indexOf("trackE")!=-1){
	    System.err.print("Parametrizing trackE ... ");
	    this.J = new Interpolate(1000, e_low, e_hi, this, 5, false, false, true, 5, false, false, false);
	    this.jt=true;
	    System.err.println("done");
	}

	s.jt=false;
	if(w.indexOf("trackX")!=-1){
	    System.err.print("Parametrizing trackX ... ");
	    s.J = new Interpolate(1000, e_low, e_hi, s, 5, false, false, true, 5, false, false, false);
	    s.jt=true;
	    System.err.println("done");
	}

	p.jt=false;
	if(w.indexOf("trackT")!=-1){
	    System.err.print("Parametrizing trackT ... ");
	    p.J = new Interpolate(1000, e_low, e_hi, p, 5, false, false, true, 5, false, false, false);
	    p.jt=true;
	    System.err.println("done");
	}

	S.jt=false;
	if(w.indexOf("contiC")!=-1){
	    System.err.print("Parametrizing contiC ... ");
	    S.J = new Interpolate(200, -5, 5, S, 5, true, false, false, 5, true, false, false);
	    S.jt=true;
	    System.err.println("done");
	}

	System.err.println("Finished parametrizations");
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * call this routine to enable interpolations and their save and reread. To enable everything, set w="all"
     */

    public void interpolate(String w, String filename){
	String name;
	boolean flag;
	name=((m.name+"_"+w).toLowerCase()).replace(' ', '-');
	name=filename+"."+name+"_"+m.ecut+"_"+m.vcut+".data";
	do {
	    flag=false;
	    try{
		Output.open(name);
		interpolate(w);
		Output.close();
	    }catch (mmcException error){
		flag=true;
		Output.delete(name);
	    }
	} while(flag);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * returns the value of the tracking integral
     */

    public double getpr(double ei, double rnd){
	if(jt) return Math.max(J.interpolate(ei), 0);
	else return I.integrateWithLog(ei, p.low, this, -rnd);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * final energy, corresponding to the given value rnd of the tracking integral
     */

    public double getef(double ei, double rndMin, double rnd){
	if(jt) return Math.min(Math.max(J.findLimit(rndMin-rnd), p.low), ei);
	else return I.getUpperLimit();
    }

    //----------------------------------------------------------------------------------------------------//

    public Interpolate J;
    public boolean jt=false;

    /**
     * 1d parametrization - interface to Interpolate
     */

    public double functionInt(double e){
	return I.integrateWithLog(e, p.low, this);
    }

}
